#include <stdlib.h>
#include <string.h>
#include <signal.h>

#ifdef WIN32
#include <windows.h>
#endif

#include "grind.h"

int interrupt = false;

int mode_opt = ALPHA, mode_none = ALPHA;

FILE *cinput;
char uline[LINE_LENGTH] = {'\0'};
char cline[LINE_LENGTH] = {'\0'};
char *cbegin = uline;
char *cend = uline;
int margc = 0;
char *margv[MAX_ARG] = {NULL};
int pargc = 0;
char **pargv = NULL;
char *modelname = NULL;
int assign = 0;
int *all_types;
_par *pars = NULL;
_var *vars = NULL;
_expr *exprs = NULL;
/*  format specifiers */
char dformat[] = " %#- 11.4g";
char lformat[] = " %#- 11.4g";
char vformat[] = " %#- 11.4g";
char aformat[] = " % 11.4g";
char sformat[] = " %-11s";

boolean *outputbuf = NULL;

/*  can't use 'const int' as size-specifier in arry declaration :( */
double *istate = NULL, *nstate = NULL;
int rundatalength = 0;
double **rundata = NULL;

extern double all_values[];
double * const fstate = all_values;


_axis axis[6] = {{NULL,0,0.0,1.0,false,3},   /*  x-axis */
		 {NULL,-1,0.0,1.0,false,3},  /*  y-axis */
		 {NULL,-1,0.0,1.0,false,3},  /*  z-axis */
		 {NULL,0,0.0,1.0,false,3},  /*  takens-axis */
		 {NULL,0,0.0,1.0,false,3},   /*  vertical axis of timeplot */
		 {NULL,0,0.0,100.0,false,5}};/*  horizontal axis of timeplot */


const char *itors[]={"row4a", "euler", "Runge Kutta"};

/*  used in odeint; if kmax == 0 none of these variables will be used */
int kmax = 0, kount = 0;
double *xp = NULL, **yp = NULL, dxsav = 0.0;

void handle_sigint (int);

int Initialize(int argc, char *argv[]) {
  int i, len;
  extern int itor; /*  specifies iterator (integ.c) */

  /*  set handlers for signals */
  /* the neat way; test whether previous handler was 'ignore', if so reset */
  if ( SIG_IGN == signal(SIGINT, handle_sigint))
    signal(SIGINT, SIG_IGN);

  /*  set pointers to original program argument */
  pargc = argc;
  pargv = argv;

  /*  length of the name of the executable - ".exe" */
  len = strlen(pargv[0])-3;
  modelname = ccalloc(len, "'modelname' in Initialize()");
  strncpy(modelname, pargv[0], len-1);
  modelname[len-1] = '\0';

  /*  initialize 'all_types' array */
  all_types = icalloc(nnames, "'all_types' in Initialize()");
  for (i=0; i<nvar; i++)
    all_types[i]=VARIABLE;
  for ( ;i<nvar+nexpr; i++)
    all_types[i]=EXPRESSION;
  for ( ;i<nvar+nexpr+npar; i++)
    all_types[i]=PARAMETER;

  /*  initialize array of parameters */
  if ((pars = (_par*)malloc(sizeof(_par)*npar)) == NULL) {
    fprintf(stderr,"Init: error in memory allocation of 'pars'\n");
    exit (1);
  }
  for (i=0; i<npar; i++) {
    pars[i].name = all_names[nvar + nexpr + i];
    pars[i].value = 0.0;
    pars[i].init = false;
    pars[i].noise = 0;
    /*     pars[i].mean = 0.0; */
    pars[i].sd = 0.0;
    pars[i].pos = false;
  }

  /*  ... ad 'vars' array */
  if ((vars = (_var*)malloc(sizeof(_var)*nvar)) == NULL) {
    fprintf(stderr,"Init: error in memory allocation of 'vars'\n");
    exit (1);
  }
  for (i=0; i<nvar; i++) {
    vars[i].name = all_names[i];
    vars[i].init = false;
  }

  /*  ... and 'expr' array */
  if ((exprs = (_expr*)malloc(sizeof(_expr)*nexpr)) == NULL) {
    fprintf(stderr,"Init: error in memory allocation of 'exprs'\n");
    exit (1);
  }
  for (i=0; i<nexpr; i++) {
    exprs[i].name = all_names[nvar + i];
    exprs[i].init = false;
  }

  /*  allocate memory for the initial and newton state */
  istate = dcalloc(nvar, "'istate' in Initialize()");
  nstate = dcalloc(nvar, "'nstate' in Initialize()");

  /*  allocate memory for the buffer which holds the trajectory data */
  /*  here we only allocate pointers to buffers, for each var & expr & par */
  /*  in 'integ()' we realloc necessary memory, */
  /*  only for those that are in the outputbuffer */
  rundata = ddcalloc(1+nnames,"'rundata' in Initialize()");
  /* allocation of rundata[i] is done in 'integ() (integ.c)' */

  /*  outputbuf keeps track for which var's/expr's/par's output is generated */
  if (!outputbuf) {
    outputbuf = bcalloc(nnames, "'outputbuf' in Initialize");
    /*  by default output is only generated for the variables */
    for (i=0; i<nvar; i++)
      outputbuf[i] = true;
    for (i=nvar; i<nnames; i++)
      outputbuf[i] = false;
  }

  /*  check the number of variables and set 'axis' variable to sensible */
  /*  default values */
  for (i=0; i<3 && i<nvar; i++) {
    axis[i].label = storeString(axis[i].label, all_names[i]);
    axis[i].var = i;
  }
  for (i=nvar; i<3; i++) {
    axis[i].label = storeString(axis[i].label, all_names[0]);
    axis[i].var = 0;
  }
  axis[T_AXIS].label = storeString(axis[T_AXIS].label, all_names[0]);
  axis[V_AXIS].label = storeString(axis[V_AXIS].label,"");
  axis[H_AXIS].label = storeString(axis[H_AXIS].label,"t");

  /*  set default input */
  cinput = stdin;

  /*  check for the existence of a jacobian */
  if (!jacob_opt) {
    itor = RK;
    gr_warning("No Jacobian: switched to Runge Kutta integration\n");
  }

  return 1;
}

int GrindMainLoop(boolean interactive) {
  while (1) {
    if (USE_GET_LINE && !interactive && Interrupt())
      return 0;
    interrupt = false;
    if (EOF == GetCommand(interactive))
      return 0;
  }
}

void handle_sigint (int signum) {

  if (!interrupt)
    printf("\nReceived user interrupt; hoping for the best\n");

  /* reenable this handler for SIGINT */
  /* the neat way; test whether previous handler was 'ignore', if so reset */
  
  if (SIG_IGN == signal(SIGINT, handle_sigint))
    signal(SIGINT, SIG_IGN);

  interrupt = true;
  return;
}

double grind_value(char *name) {
  int i;
  if ( (i=nameIndex(name)) >=0 && i < nnames) {
    if (calcul_opt && i<nvar)
      return istate[i];
    else
      return all_values[i];
  }
  return 0.;
}
      
int Interrupt(){
  return interrupt;
}
