#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>

#include "grind.h"

static double 
  *state = NULL,             /*  a local state vector */
  *rates = NULL;             /*  a local derivatives vector */
static int precision = 200;  /*  density of calculations */
static double **rsign = NULL;    /*  will store the derivatives of 'var' */

int plane(int, int, int, int, double);
int memalloc();

int shade(int var, int ax, int shint) {
  double dist = 1.0/(double)shint;
  double begin = dist/2.0;
  int i;

  if (Interrupt())
    return 0;

  if (DIM3 != mode_opt)
    return 0;

  /*  allocate memory when first time */
  memalloc();

  for (i=0; i<shint; i++, begin+=dist) {
    if (X_AXIS == ax)
      plane(var, 1, 2, 0, begin);
    else if (Y_AXIS == ax)
      plane(var, 2, 0, 1, begin);
    else
      plane(var, 0, 1, 2, begin);
  }
  return 1;
}  

int nullcline(int var) {
  if (Interrupt())
    return 0;

  /*  allocate memory when first time */
  memalloc();

  
  if (mode_opt == DIM2)
    /*  draw nullclines in std x-y plane for z=0. */
    plane(var, 0, 1, 2, 0.0);
  if (mode_opt == DIM3) {
    /*  x-y plane, z = 0 */
    plane(var, 0, 1, 2, 0.0);
    /*  y-z plane, x = 0 */
    plane(var, 2, 1, 0, 0.0);
    /*  z-x plane, y = 0 */
    plane(var, 0, 2, 1, 0.0);
    /*  x-y plane, z = 1 */
    plane(var, 0, 1, 2, 1.0);
    /*  y-z plane, x = 1 */
    plane(var, 2, 1, 0, 1.0);
    /*  z-x plane, y = 1 */
    plane(var, 0, 2, 1, 1.0);
  }
  
  return 1;
}

int plane(int var, int ax0, int ax1, int ax2, double value2) {
  /*  var is the variable for which nulllcline is computed */
  /*  ax0 and ax1 define the plane in which the nullcline is calculated, */
  /*  value2 specifies the value along the 3rd axis */

  /*  'scale' rescales loop-iterators between 0. and 1. */
  double scale = 1./precision;
  /*  p[] will hold the coordinates to be drawn */
  double p[3];
  /*  iterators */
  int i, j, col;
  /*  a flag which checks whether a plane is in fact a nullplane */
  boolean all_null = true;
  boolean N,S,E,W;
  double dNW, dNE, dSE, dSW, absNW, absNE, absSE, absSW;
  double Npos=0, Epos=0, Spos=0, Wpos=0;

  if (Interrupt())
    return 0;

  /*  initialize 'state' with initial values */
  for (i=0; i<nvar; i++)
    state[i] = istate[i];

  /*  temporary store the values of all parameters and expressions */
  /*  at the end we'll restore these values again */
  storeaxis(true);

  /*  set var, expr, or par, corresponding to 3rd axis, to max or min value */
  if (mode_opt == DIM3)
    setax(state, value2, ax2);

  /*  loop over ax0 and ax1 (with 'precision' number of steps) and calculate */
  /*  where the derivative of 'var' is zero. store this info in 'null[][]' */
  /*  keep track whether all positions in null[][] are zero, using 'all_null' */
  for (i=0; i <= precision; i++) {
    /*  adjust state variable for var v0 */
    setax(state, scale*i, ax0);
    for (j=0; j <= precision; j++) {
      if (Interrupt()) {
	storeaxis(false);
	return 0;
      }
	
      /*  adjust state variable for var v0 */
      setax(state, scale*j, ax1);
      /*  given 'state' get derivatives */
      fcn(state,rates);
      rsign[i][j] = rates[var]; /* Store sign of derivatives of "var" */
      all_null &= (0.0 == rsign[i][j]);  
    }
  }

  /*  start drawing with a new color */;
  col = var + 2;
  addGrcall(COLOR, (void*)&col, 0., 0., 0.);
  
  /*  here we fill p[]; ax0, ax1, and ax2 define the coord system to draw in */
  p[ax2] = value2;
  if (all_null) {
    p[ax0] = p[ax1] = 0.0;
    addGrcall(MOVE, NULL, p[0], p[1], p[2]);
    p[ax0] = precision*scale; 
    addGrcall(DRAW, NULL, p[0], p[1], p[2]);
    p[ax1] = precision*scale;
    addGrcall(DRAW, NULL, p[0], p[1], p[2]);
    p[ax0] = 0.0;
    addGrcall(DRAW, NULL, p[0], p[1], p[2]);
    p[ax1] = 0.0;
    addGrcall(DRAW, NULL, p[0], p[1], p[2]);
  }
  else {
    for (i=0; i < precision; i++) 
    for (j=0; j < precision; j++) {
      if (Interrupt()) {
	storeaxis(false);
	return 0;
      }
      N = S = E = W = false;
      dSW = rsign[i][j];     absSW = fabs(dSW);
      dNE = rsign[i+1][j+1]; absNE = fabs(dNE);
      dSE = rsign[i+1][j];   absSE = fabs(dSE);
      dNW = rsign[i][j+1];   absNW = fabs(dNW);

      if (dNW*dNE <= 0.0) {
	N = true; 
	if (0.0 == dNW+dNE) Npos = 0.5;
	else Npos = absNW/(absNW+absNE);
      }
      if (dSW*dSE <= 0.0) {
	S = true;
	if (0.0 == dSW+dSE) Spos = 0.5;
	else Spos = absSW/(absSW+absSE);
      }
      if (dNE*dSE <= 0.0) {
	E = true; 
	if (0.0 == dNE+dSE) Epos = 0.5;
	else Epos = absSE/(absSE+absNE);
      } 
      if (dNW*dSW <= 0.0) {
	W = true; 
	if (0.0 == dNW+dSW) Wpos = 0.5;
	else Wpos = absSW/(absNW+absSW);
      }

      if (S && N) {
	p[ax0] = (i+Spos)*scale; p[ax1] = j*scale;
	addGrcall(MOVE, NULL, p[0], p[1], p[2]);
	p[ax0] = (i+Npos)*scale; p[ax1] = (j+1)*scale;
	addGrcall(DRAW, NULL, p[0], p[1], p[2]);
      }
      if (W && E) {
	p[ax1] = (j+Wpos)*scale; p[ax0] = i*scale;
	addGrcall(MOVE, NULL, p[0], p[1], p[2]);
	p[ax1] = (j+Epos)*scale; p[ax0] = (i+1)*scale;
	addGrcall(DRAW, NULL, p[0], p[1], p[2]);
      }
      if (S && E && !N && !W) {
	p[ax0] = (i+Spos)*scale; p[ax1] = j*scale;
	addGrcall(MOVE, NULL, p[0], p[1], p[2]);
	p[ax0] = (i+1)*scale; p[ax1] = (j+Epos)*scale;
	addGrcall(DRAW, NULL, p[0], p[1], p[2]);
      }      
      if (S && W && !N && !E) {
	p[ax0] = (i+Spos)*scale; p[ax1] = j*scale;
	addGrcall(MOVE, NULL, p[0], p[1], p[2]);
	p[ax0] = i*scale; p[ax1] = (j+Wpos)*scale;
	addGrcall(DRAW, NULL, p[0], p[1], p[2]);
      }
      if (N && E && !S && !W) {
	p[ax0] = (i+Npos)*scale; p[ax1] = (j+1)*scale;
	addGrcall(MOVE, NULL, p[0], p[1], p[2]);
	p[ax0] = (i+1)*scale; p[ax1] = (j+Epos)*scale;
	addGrcall(DRAW, NULL, p[0], p[1], p[2]);
      }   
      if (N && W && !S && !E) {
	p[ax0] = (i+Npos)*scale; p[ax1] = (j+1)*scale;
	addGrcall(MOVE, NULL, p[0], p[1], p[2]);
	p[ax0] = i*scale; p[ax1] = (j+Wpos)*scale;
	addGrcall(DRAW, NULL, p[0], p[1], p[2]);
      }     
    }
  }
  /*  restore parameters and expressions to original values */
  storeaxis(false);
  return 1;
}

int memalloc() {
  int i;

  /*  first time allocation */
  if(state == NULL)
    state = dcalloc(nvar, "'state' in memalloc() (nullcline.c)");
  /*  first time allocation */
  if(rates == NULL)
    rates = dcalloc(nvar, "'rates' in memalloc() (nullcline.c)");
  /*  first time allocation */
  if (rsign == NULL) {
    rsign = ddcalloc(precision+1, "'rsign' in memalloc()");
    rsign[0] = dcalloc((precision+1)*(precision+1), 
		       "'rsign[0]' in memalloc()");
    for (i=1; i<precision+1; i++)
      rsign[i] = rsign[i-1]+precision+1;
  }
  return precision;
}

